# -*- coding: utf-8 -*-

# Copyright (C) 2007-2013 Avencall
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

import logging
import random
import string
import time

from xivo_dao import agent_dao
from xivo_dao import group_dao
from xivo_dao import incall_dao
from xivo_dao import user_dao
from xivo_dao import queue_dao

ALPHANUMS = string.uppercase + string.lowercase + string.digits

logger = logging.getLogger('AMI_1.8')


class AMI_1_8(object):

    userevents = ('Feature',
                  'dialplan2cti',
                  'Agent',
                  'User',
                  'Queue',
                  'Group',
                  'Did',)

    def __init__(self, cti_server, innerdata, interface_ami):
        self._ctiserver = cti_server
        self.innerdata = innerdata
        self.interface_ami = interface_ami

    def ami_newchannel(self, event):
        channel = event['Channel']
        state = event['ChannelState']
        state_description = event['ChannelStateDesc']
        context = event['Context']
        unique_id = event['Uniqueid']

        self.innerdata.newchannel(channel, context, state, state_description, unique_id)

    def ami_hangup(self, event):
        channel = event['Channel']
        #  0 - Unknown
        #  3 - No route to destination
        # 16 - Normal Clearing
        # 17 - User busy (see Orig #5)
        # 18 - No user responding (see Orig #1)
        # 19 - User alerting, no answer (see Orig #8, Orig #3, Orig #1 (soft hup))
        # 21 - Call rejected (attempting *8 when noone to intercept)
        # 24 - "lost" Call suspended
        # 27 - Destination out of order
        # 28 - Invalid number format (incomplete number)
        # 34 - Circuit/channel congestion
        self.innerdata.sheetsend('hangup', channel)
        self.innerdata.hangup(channel)

    def ami_dial(self, event):
        channel = event['Channel']
        subevent = event['SubEvent']
        if subevent == 'Begin' and 'Destination' in event:
            destination = event['Destination']
            if channel in self.innerdata.channels:
                try:
                    self.innerdata.channels[channel].set_extra_data('xivo', 'desttype', 'user')
                    phone = self.innerdata.xod_config['phones'].find_phone_by_channel(destination)
                    if phone:
                        self.innerdata.channels[channel].set_extra_data('xivo', 'destid', str(phone['iduserfeatures']))
                except LookupError:
                    logger.exception('Could not set user id for dial')
                self.innerdata.channels[channel].properties['direction'] = 'out'
                self.innerdata.channels[channel].properties['commstatus'] = 'calling'
                self.innerdata.channels[channel].properties['timestamp'] = time.time()
                self.innerdata.setpeerchannel(channel, destination)
                self.innerdata.update(channel)
            if destination in self.innerdata.channels:
                self.innerdata.channels[destination].properties['direction'] = 'in'
                self.innerdata.channels[destination].properties['commstatus'] = 'ringing'
                self.innerdata.channels[destination].properties['timestamp'] = time.time()
                self.innerdata.setpeerchannel(destination, channel)
                self.innerdata.update(destination)
            self.innerdata.sheetsend('dial', channel)

    def ami_extensionstatus(self, event):
        self.innerdata.updatehint(event['Hint'], event['Status'])

    def ami_bridge(self, event):
        if event['Bridgestate'] != 'Link':
            return

        channel_1 = self.innerdata.channels.get(event['Channel1'])
        channel_2 = self.innerdata.channels.get(event['Channel2'])

        if channel_1:
            self._update_connected_channel(channel_1, channel_2)
            channel_1.properties['commstatus'] = 'linked-caller'

        if channel_2:
            self._update_connected_channel(channel_2, channel_1)
            channel_2.properties['commstatus'] = 'linked-called'

        if 'agentcallback' in channel_1.channel or 'agentcallback' in channel_2.channel:
            logger.debug('A call has been bridged to an agent, not sending the link event now')
        else:
            self._trigger_link_event(channel_2)

    def _update_connected_channel(self, channel, peer_channel):
        channel.properties.update({
            'talkingto_kind': 'channel',
            'talkingto_id': peer_channel.channel,
            'timestamp': time.time(),
        })
        self.innerdata.setpeerchannel(channel.channel, peer_channel.channel)
        self.innerdata.update(channel.channel)

    def _trigger_link_event(self, channel):
        phone = self.innerdata.xod_config['phones'].find_phone_by_channel(channel.channel)
        if not phone:
            return

        channel.set_extra_data('xivo', 'desttype', 'user')
        channel.set_extra_data('xivo', 'destid', str(phone['iduserfeatures']))

        self.innerdata.sheetsend('link', channel.channel)

    def ami_unlink(self, event):
        self.innerdata.sheetsend('unlink', event['Channel1'])

    def ami_masquerade(self, event):
        original = event['Original']
        clone = event['Clone']
        self.innerdata.masquerade(original, clone)

    def ami_channelupdate(self, event):
        # could be especially useful when there is a trunk : links callno-remote and callno-local
        # when the call is outgoing, one never receives the callno-remote
        channeltype = event['Channeltype']
        if channeltype == 'IAX2':
            logger.info('ami_channelupdate %s : %s - %s : %s',
                        channeltype, event['IAX2-callno-local'],
                        event['IAX2-callno-remote'],
                        event)

    def ami_originateresponse(self, event):
        channel = event['Channel']
        actionid = event.get('ActionID')
        reason = event['Reason']
        # reasons ...
        # 4 : Success
        # 0 : 1st phone unregistered
        # 1 : CLI 'channel request hangup' on the 1st phone's channel
        # 5 : 1st phone rejected the call (reject button or all lines busy)
        # 8 : 1st phone did not answer early enough
        if actionid in self.interface_ami.originate_actionids:
            properties = self.interface_ami.originate_actionids.pop(actionid)
            request = properties.get('request')
            cn = request.get('requester')
            try:
                cn.reply({'class': 'ipbxcommand',
                          'command': request.get('ipbxcommand'),
                          'replyid': request.get('commandid'),
                          'channel': channel,
                          'originatereason': reason})
            except Exception:
                # when requester is not connected any more ...
                pass

    def ami_parkedcall(self, event):
        channel = event['Channel']
        exten = event['Exten']
        parkinglot = event['Parkinglot']
        if parkinglot.startswith('parkinglot_'):
            parkinglot = '_'.join(parkinglot.split('_')[1:])
        parkingevent = {
            'parker': event.pop('From'),
            'parked': channel,
            'exten': exten,
            'cid_name': event.pop('CallerIDName'),
            'cid_num': event.pop('CallerIDNum'),
            'parktime': time.time(),
            }
        if channel in self.innerdata.channels:
            self.innerdata.channels[channel].setparking(exten, parkinglot)

    def ami_unparkedcall(self, event):
        channel = event['Channel']
        if channel in self.innerdata.channels:
            self.innerdata.channels[channel].unsetparking()

    def ami_parkedcalltimeout(self, event):
        channel = event['Channel']
        if channel in self.innerdata.channels:
            self.innerdata.channels[channel].unsetparking()

    def userevent_agent(self, chanprops, event):
        agent_id = int(event['XIVO_AGENT_ID'])
        try:
            agent_number = agent_dao.agent_number(agent_id)

            # chanprops is the ";2" channel of a "Local channel", but we need
            # to set the extra_data information on the ";1" channel
            semicolon1_channel = chanprops.channel[:-1] + '1'
            semicolon1_chanprops = self.innerdata.channels[semicolon1_channel]
            semicolon1_chanprops.set_extra_data('xivo', 'agentnumber', agent_number)
        except Exception:
            logger.warning('Could not set extra data in userevent agent', exc_info=True)

    def userevent_user(self, chanprops, event):
        xivo_userid = event.get('XIVO_USERID')
        userprops = self.innerdata.xod_config.get('users').keeplist.get(xivo_userid)
        xivo_srcnum = event.get('XIVO_SRCNUM')
        destination_user_id = int(event['XIVO_DSTID'])
        destination_name, destination_number = user_dao.get_name_number(destination_user_id)
        if userprops is not None:
            usersummary_src = {'fullname': userprops.get('fullname'),
                               'phonenumber': xivo_srcnum}
        else:
            usersummary_src = {'fullname': xivo_srcnum,
                               'phonenumber': xivo_srcnum}

        chanprops.set_extra_data('xivo', 'desttype', 'user')
        chanprops.set_extra_data('xivo', 'destid', destination_user_id)
        chanprops.set_extra_data('xivo', 'userid', xivo_userid)
        chanprops.set_extra_data('xivo', 'origin', event.get('XIVO_CALLORIGIN', 'internal'))
        chanprops.set_extra_data('xivo', 'direction', 'internal')
        if not chanprops.has_extra_data('xivo', 'calleridname'):
            chanprops.set_extra_data('xivo', 'calleridname', usersummary_src.get('fullname'))
        if not chanprops.has_extra_data('xivo', 'calleridnum'):
            chanprops.set_extra_data('xivo', 'calleridnum', usersummary_src.get('phonenumber'))
        chanprops.set_extra_data('xivo', 'calledidnum', destination_number)
        chanprops.set_extra_data('xivo', 'calledidname', destination_name)

    def userevent_queue(self, chanprops, event):
        callerid_name = event.get('XIVO_CALLERIDNAME')
        callerid_number = event.get('XIVO_CALLERIDNUMBER')

        queue_id = int(event['XIVO_DSTID'])
        queue_name, queue_number = queue_dao.get_display_name_number(queue_id)

        chanprops.set_extra_data('xivo', 'desttype', 'queue')
        chanprops.set_extra_data('xivo', 'destid', queue_id)
        chanprops.set_extra_data('xivo', 'calledidname', queue_name)
        chanprops.set_extra_data('xivo', 'queuename', queue_name)
        chanprops.set_extra_data('xivo', 'calledidnum', queue_number)
        if not chanprops.has_extra_data('xivo', 'calleridname'):
            chanprops.set_extra_data('xivo', 'calleridname', callerid_name)
        if not chanprops.has_extra_data('xivo', 'calleridnum'):
            chanprops.set_extra_data('xivo', 'calleridnum', callerid_number)

        self.innerdata.sheetsend('dial', chanprops.channel)

    def userevent_group(self, chanprops, event):
        group_id = int(event['XIVO_DSTID'])
        group_name, group_number = group_dao.get_name_number(group_id)

        chanprops.set_extra_data('xivo', 'desttype', 'group')
        chanprops.set_extra_data('xivo', 'destid', group_id)
        chanprops.set_extra_data('xivo', 'calledidname', group_name)
        chanprops.set_extra_data('xivo', 'calledidnum', group_number)

        self.innerdata.sheetsend('dial', chanprops.channel)

    def userevent_did(self, chanprops, event):
        calleridnum = event.get('XIVO_SRCNUM')
        calleridname = event.get('XIVO_SRCNAME')
        calleridton = event.get('XIVO_SRCTON')
        calleridrdnis = event.get('XIVO_SRCRDNIS')
        didnumber = event.get('XIVO_EXTENPATTERN')

        chanprops.set_extra_data('xivo', 'origin', 'did')
        chanprops.set_extra_data('xivo', 'direction', 'incoming')
        chanprops.set_extra_data('xivo', 'did', didnumber)
        chanprops.set_extra_data('xivo', 'calleridnum', calleridnum)
        chanprops.set_extra_data('xivo', 'calleridname', calleridname)
        chanprops.set_extra_data('xivo', 'calleridrdnis', calleridrdnis)
        chanprops.set_extra_data('xivo', 'calleridton', calleridton)
        incall = incall_dao.get_by_exten(didnumber)
        if incall:
            chanprops.set_extra_data('xivo', 'desttype', incall.action)
            chanprops.set_extra_data('xivo', 'destid', incall.actionarg1)
        self.innerdata.sheetsend('incomingdid', chanprops.channel)

    def userevent_feature(self, chanprops, ev):
        reply = {}
        if 'XIVO_USERID' in ev and 'Function' in ev and 'Status' in ev:
            userid = ev['XIVO_USERID']
            status = int(ev['Status']) != 0
            user = self.innerdata.xod_config['users'].keeplist[userid]
            fn = ev['Function']
            if 'dnd' in fn:
                user['enablednd'] = status
            if 'callrecord' in fn:
                user['callrecord'] = status
            if 'incallfilter' in fn:
                user['incallfilter'] = status
            if 'unc' == fn:
                fn = 'enableunc'
                user[fn] = status
                user['destunc'] = ev.get('Value')
            if 'rna' == fn:
                fn = 'enablerna'
                user[fn] = status
                user['destrna'] = ev.get('Value')
            if 'busy' == fn:
                fn = 'enablebusy'
                user[fn] = status
                user['destbusy'] = ev.get('Value')
            event = {'class': 'getlist',
                     'listname': 'users',
                     'function': 'updateconfig',
                     'tipbxid': self._ctiserver.myipbxid,
                     'tid': userid,
                     'config': user}
            self._ctiserver.send_cti_event(event)
        return reply

    def userevent_dialplan2cti(self, chanprops, event):
        # why "UserEvent + dialplan2cti" and not "Newexten + Set" ?
        # - more selective
        # - variables declarations are not always done with Set (Read(), AGI(), ...)
        # - if there is a need for extra useful data (XIVO_USERID, ...)
        # - (future ?) multiple settings at once
        cti_varname = event.get('VARIABLE')
        dp_value = event.get('VALUE')
        chanprops.set_extra_data('dp', cti_varname, dp_value)

    def ami_userevent(self, event):
        eventname = event['UserEvent']
        channel = event.get('CHANNEL')
        # syntax in dialplan : exten = xx,n,UserEvent(myevent,var1: ${var1},var2: ${var2})
        if channel in self.innerdata.channels:
            if eventname in self.userevents:
                methodname = 'userevent_%s' % eventname.lower()
                if hasattr(self, methodname):
                    chanprops = self.innerdata.channels[channel]
                    getattr(self, methodname)(chanprops, event)

    def ami_messagewaiting(self, event):
        try:
            full_mailbox = event['Mailbox']
            for mailbox_id, mailbox in self.innerdata.xod_config['voicemails'].keeplist.iteritems():
                if full_mailbox == mailbox['fullmailbox']:
                    previous_status = self.innerdata.xod_status['voicemails'][mailbox_id]
                    old = event.get('Old') or previous_status['old']
                    new = event.get('New') or previous_status['new']
                    waiting = event.get('Waiting') or previous_status['waiting']
                    logger.info("Voicemail event received. mailbox:%s new:%s old:%s waiting:%s", full_mailbox, new, old, waiting)
                    self.innerdata.voicemailupdate(full_mailbox, new, old, waiting)
            if 'Old' not in event and 'New' not in event:
                logger.info("Voicemail event did not contain 'old' and 'new' count. retrieving mailbox count")
                params = {'mode': 'vmupdate',
                          'amicommand': 'mailboxcount',
                          'amiargs': full_mailbox.split('@')}
                actionid = ''.join(random.sample(ALPHANUMS, 10))
                self.interface_ami.execute_and_track(actionid, params)
        except KeyError:
            logger.warning('ami_messagewaiting Failed to update mailbox')

    def ami_coreshowchannel(self, event):
        channel = event['Channel']
        context = event['Context']
        application = event['Application']
        bridgedchannel = event['BridgedChannel']
        state = event['ChannelState']
        state_description = event['ChannelStateDesc']
        timestamp_start = self.timeconvert(event['Duration'])
        unique_id = event['UniqueID']

        self.innerdata.newchannel(channel, context, state, state_description, unique_id)
        channelstruct = self.innerdata.channels[channel]

        channelstruct.properties['timestamp'] = timestamp_start
        if application == 'Dial':
            channelstruct.properties['direction'] = 'out'
            if state == '6':
                channelstruct.properties['commstatus'] = 'linked-caller'
            elif state == '4':
                channelstruct.properties['commstatus'] = 'calling'
        elif application == 'AppDial':
            channelstruct.properties['direction'] = 'in'
            if state == '6':
                channelstruct.properties['commstatus'] = 'linked-called'
            elif state == '5':
                channelstruct.properties['commstatus'] = 'ringing'

        if state == '6' and bridgedchannel:
            self.innerdata.newchannel(bridgedchannel, context, state, state_description, unique_id)
            self.innerdata.setpeerchannel(channel, bridgedchannel)
            self.innerdata.setpeerchannel(bridgedchannel, channel)

    def ami_listdialplan(self, event):
        extension = event.get('Extension')
        if extension and extension.isdigit() and event['Priority'] == '1':
            actionid = 'exten:%s' % ''.join(random.sample(ALPHANUMS, 10))
            params = {'mode': 'extension',
                      'amicommand': 'sendextensionstate',
                      'amiargs': (extension, event['Context'])}
            self.interface_ami.execute_and_track(actionid, params)

    def ami_voicemailuserentry(self, event):
        fullmailbox = '%s@%s' % (event['VoiceMailbox'], event['VMContext'])
        # only NewMessageCount here - OldMessageCount only when IMAP compiled
        # relation to Old/New/Waiting in MessageWaiting UserEvent ?
        self.innerdata.voicemailupdate(fullmailbox, event['NewMessageCount'])

    def amiresponse_extensionstatus(self, event):
        hint = event.get('Hint')
        if hint:
            self.innerdata.updatehint(hint, event['Status'])

    @staticmethod
    def timeconvert(duration):
        if ':' in duration:
            # like in core show channels output
            vdur = duration.split(':')
            duration_secs = int(vdur[0]) * 3600 + int(vdur[1]) * 60 + int(vdur[2])
        else:
            # like in queue entry output
            duration_secs = int(duration)
        timestamp_start = time.time() - duration_secs
        return timestamp_start
