# -*- coding: utf-8 -*-

# Copyright (C) 2013 Avencall
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>


import logging
import os
import Queue
import select
import signal
import socket
import ssl
import sys
import time
import threading

from xivo import daemonize
from xivo_cti import cti_config
from xivo_cti import dao
from xivo_cti import message_hook
from xivo_cti.ami import ami_callback_handler
from xivo_cti import channel_updater
from xivo_cti.client_connection import ClientConnection
from xivo_cti.cti.commands.agent_login import AgentLogin
from xivo_cti.cti.commands.agent_logout import AgentLogout
from xivo_cti.cti.commands.answer import Answer
from xivo_cti.cti.commands.dial import Dial
from xivo_cti.cti.commands.attended_transfer import AttendedTransfer
from xivo_cti.cti.commands.direct_transfer import DirectTransfer
from xivo_cti.cti.commands.cancel_transfer import CancelTransfer
from xivo_cti.cti.commands.complete_transfer import CompleteTransfer
from xivo_cti.cti.commands.hangup import Hangup
from xivo_cti.cti.commands.listen import Listen
from xivo_cti.cti.commands.logout import Logout
from xivo_cti.cti.commands.queue_add import QueueAdd
from xivo_cti.cti.commands.queue_pause import QueuePause
from xivo_cti.cti.commands.queue_remove import QueueRemove
from xivo_cti.cti.commands.queue_unpause import QueueUnPause
from xivo_cti.cti.commands.subscribe import SubscribeCurrentCalls, \
    SubscribeMeetmeUpdate, SubscribeQueueEntryUpdate
from xivo_cti.cti.commands.subscribetoqueuesstats import SubscribeToQueuesStats
from xivo_cti.cti.commands.hold_switchboard import HoldSwitchboard
from xivo_cti.cti.commands.unhold_switchboard import UnholdSwitchboard
from xivo_cti.cti.commands.set_forward import DisableBusyForward, \
    DisableNoAnswerForward, DisableUnconditionalForward, EnableBusyForward, \
    EnableNoAnswerForward, EnableUnconditionalForward
from xivo_cti.cti.commands.set_user_service import DisableDND, DisableFilter, EnableDND, EnableFilter, \
    EnableRecording, DisableRecording
from xivo_cti.services.funckey import manager as funckey_manager
from xivo_cti.interfaces import interface_cti
from xivo_cti.interfaces import interface_info
from xivo_cti.interfaces import interface_webi
from xivo_cti.interfaces.interfaces import DisconnectCause
from xivo_cti.queue_logger import QueueLogger
from xivo_cti.services import queue_entry_manager
from xivo_cti.services.agent.status import AgentStatus
from xivo_cti.services.meetme import service_manager as meetme_service_manager_module
from xivo_cti.statistics import queue_statistics_manager
from xivo_cti.statistics import queue_statistics_producer
from xivo_cti.ioc.context import context

logger = logging.getLogger('main')


class CTIServer(object):

    servername = 'XiVO CTI Server'

    def __init__(self, config):
        self.start_time = time.time()
        self.mycti = {}
        self.myipbxid = 'xivo'
        self.interface_ami = None
        self.timeout_queue = None
        self.update_config_list = []
        self.pipe_queued_threads = os.pipe()
        self._config = config
        self._cti_events = Queue.Queue()

    def _set_signal_handlers(self):
        signal.signal(signal.SIGINT, self._sighandler)
        signal.signal(signal.SIGTERM, self._sighandler)
        signal.signal(signal.SIGHUP, self._sighandler_reload)

    def _sighandler(self, signum, frame):
        logger.warning('(sighandler) signal %s lineno %s (atq = %s) received : quits',
                       signum, frame.f_lineno, self.askedtoquit)
        for t in filter(lambda x: x.getName() != 'MainThread', threading.enumerate()):
            t._Thread__stop()
        self.askedtoquit = True

    def _sighandler_reload(self, signum, frame):
        logger.warning('(sighandler_reload) signal %s lineno %s (atq = %s) received : reloads',
                       signum, frame.f_lineno, self.askedtoquit)
        self.askedtoquit = False

    def _set_logger(self):
        file_handler = logging.FileHandler(cti_config.LOGFILENAME)
        file_formatter = logging.Formatter(
            '%%(asctime)s %s[%%(process)d] (%%(levelname)s) (%%(name)s): %%(message)s'
            % cti_config.DAEMONNAME)
        file_handler.setFormatter(file_formatter)
        root_logger = logging.getLogger()
        root_logger.addHandler(file_handler)
        root_logger.setLevel(logging.INFO)
        if cti_config.DEBUG_MODE:
            stream_handler = logging.StreamHandler()
            stream_formatter = logging.Formatter('%(asctime)s %(levelname)s:%(name)s: %(message)s')
            stream_handler.setFormatter(stream_formatter)
            root_logger.addHandler(stream_handler)
            root_logger.setLevel(logging.DEBUG)

    def _daemonize(self):
        if not cti_config.DEBUG_MODE:
            daemonize.daemonize()
        daemonize.lock_pidfile_or_die(cti_config.PIDFILE)

    def setup(self):
        self._set_logger()
        self._daemonize()
        QueueLogger.init()
        self._config.update()
        self.timeout_queue = Queue.Queue()
        self._set_signal_handlers()

        self.interface_ami = context.get('interface_ami')

        self._user_service_manager = context.get('user_service_manager')
        self._funckey_manager = context.get('funckey_manager')
        self._agent_service_manager = context.get('agent_service_manager')

        self._presence_service_executor = context.get('presence_service_executor')
        self._statistics_notifier = context.get('statistics_notifier')
        self._queue_statistics_producer = context.get('queue_statistics_producer')

        self._queue_member_notifier = context.get('queue_member_notifier')
        self._queue_member_updater = context.get('queue_member_updater')
        self._queue_member_manager = context.get('queue_member_manager')
        self._queue_member_cti_adapter = context.get('queue_member_cti_adapter')
        self._queue_member_cti_subscriber = context.get('queue_member_cti_subscriber')
        self._queue_member_indexer = context.get('queue_member_indexer')

        self._user_service_manager.presence_service_executor = self._presence_service_executor

        self._queue_entry_manager = context.get('queue_entry_manager')
        self._queue_statistics_manager = context.get('queue_statistics_manager')
        self._queue_entry_notifier = context.get('queue_entry_notifier')

        scheduler = context.get('scheduler')
        scheduler.setup(self.pipe_queued_threads[1])

        self._agent_availability_updater = context.get('agent_availability_updater')
        self._agent_service_cti_parser = context.get('agent_service_cti_parser')

        self._statistics_producer_initializer = context.get('statistics_producer_initializer')

        self._agent_status_manager = context.get('agent_status_manager')

        self._agent_client = context.get('agent_client')
        self._agent_client.connect('localhost')

        context.get('user_service_notifier').send_cti_event = self.send_cti_event
        context.get('user_service_notifier').ipbx_id = self.myipbxid

        queue_entry_manager.register_events()
        queue_statistics_manager.register_events()
        queue_statistics_producer.register_events()
        meetme_service_manager_module.register_callbacks()

        meetme_service_manager = context.get('meetme_service_manager')
        meetme_service_manager.initialize()

        self._register_cti_callbacks()
        self._register_ami_callbacks()
        self._register_message_hooks()

    def _register_cti_callbacks(self):
        Answer.register_callback_params(self._user_service_manager.pickup_the_phone, ['user_id'])
        Dial.register_callback_params(self._user_service_manager.call_destination,
                                      ['cti_connection', 'user_id', 'destination'])
        EnableDND.register_callback_params(self._user_service_manager.enable_dnd, ['user_id'])
        DisableDND.register_callback_params(self._user_service_manager.disable_dnd, ['user_id'])
        EnableRecording.register_callback_params(self._user_service_manager.enable_recording, ['target'])
        DisableRecording.register_callback_params(self._user_service_manager.disable_recording, ['target'])
        EnableFilter.register_callback_params(self._user_service_manager.enable_filter, ['user_id'])
        DisableFilter.register_callback_params(self._user_service_manager.disable_filter, ['user_id'])
        EnableUnconditionalForward.register_callback_params(self._user_service_manager.enable_unconditional_fwd, ['user_id', 'destination'])
        DisableUnconditionalForward.register_callback_params(self._user_service_manager.disable_unconditional_fwd, ['user_id', 'destination'])
        EnableNoAnswerForward.register_callback_params(self._user_service_manager.enable_rna_fwd, ['user_id', 'destination'])
        DisableNoAnswerForward.register_callback_params(self._user_service_manager.disable_rna_fwd, ['user_id', 'destination'])
        EnableBusyForward.register_callback_params(self._user_service_manager.enable_busy_fwd, ['user_id', 'destination'])
        DisableBusyForward.register_callback_params(self._user_service_manager.disable_busy_fwd, ['user_id', 'destination'])

        SubscribeToQueuesStats.register_callback_params(self._statistics_notifier.subscribe, ['cti_connection'])
        SubscribeToQueuesStats.register_callback_params(self._queue_statistics_producer.send_all_stats, ['cti_connection'])
        SubscribeToQueuesStats.register_callback_params(self._queue_statistics_manager.get_all_queue_summary)
        SubscribeToQueuesStats.register_callback_params(self._queue_entry_manager.publish_all_realtime_stats, ['cti_connection'])
        SubscribeQueueEntryUpdate.register_callback_params(
            self._queue_entry_notifier.subscribe, ['cti_connection', 'queue_id'])

        AgentLogin.register_callback_params(
            self._agent_service_manager.on_cti_agent_login,
            ['user_id', 'agent_id', 'agent_phone_number']
        )
        AgentLogout.register_callback_params(
            self._agent_service_manager.on_cti_agent_logout,
            ['user_id', 'agent_id']
        )
        Listen.register_callback_params(
            self._agent_service_manager.on_cti_listen,
            ['user_id', 'destination']
        )

        Logout.register_callback_params(self._user_service_manager.disconnect, ['user_id'])

        QueueAdd.register_callback_params(
            self._agent_service_cti_parser.queue_add,
            ['member', 'queue']
        )
        QueueRemove.register_callback_params(
            self._agent_service_cti_parser.queue_remove,
            ['member', 'queue']
        )
        QueuePause.register_callback_params(
            self._agent_service_cti_parser.queue_pause,
            ['member', 'queue']
        )
        QueueUnPause.register_callback_params(
            self._agent_service_cti_parser.queue_unpause,
            ['member', 'queue']
        )

        SubscribeMeetmeUpdate.register_callback_params(
            context.get('meetme_service_notifier').subscribe,
            ['cti_connection']
        )
        SubscribeCurrentCalls.register_callback_params(
            context.get('current_call_notifier').subscribe,
            ['cti_connection']
        )
        Hangup.register_callback_params(
            context.get('current_call_manager').hangup,
            ['user_id']
        )
        AttendedTransfer.register_callback_params(
            context.get('current_call_manager').attended_transfer,
            ['user_id', 'number']
        )
        DirectTransfer.register_callback_params(
            context.get('current_call_manager').direct_transfer,
            ['user_id', 'number']
        )
        CancelTransfer.register_callback_params(
            context.get('current_call_manager').cancel_transfer,
            ['user_id']
        )
        CompleteTransfer.register_callback_params(
            context.get('current_call_manager').complete_transfer,
            ['user_id']
        )
        HoldSwitchboard.register_callback_params(
            context.get('current_call_manager').switchboard_hold,
            ['user_id', 'queue_name']
        )
        UnholdSwitchboard.register_callback_params(
            context.get('current_call_manager').switchboard_unhold,
            ['user_id', 'unique_id'],
        )

    def _register_ami_callbacks(self):
        callback_handler = ami_callback_handler.AMICallbackHandler.get_instance()
        agent_status_parser = context.get('agent_status_parser')

        self._queue_member_updater.register_ami_events(callback_handler)

        callback_handler.register_callback('QueueMemberPaused', agent_status_parser.parse_ami_paused)
        callback_handler.register_callback('AgentConnect', agent_status_parser.parse_ami_acd_call_start)
        callback_handler.register_callback('AgentComplete', agent_status_parser.parse_ami_acd_call_end)
        callback_handler.register_userevent_callback('AgentLogin', agent_status_parser.parse_ami_login)
        callback_handler.register_userevent_callback('AgentLogoff', agent_status_parser.parse_ami_logout)

        current_call_parser = context.get('current_call_parser')
        current_call_parser.register_ami_events()

        callback_handler.register_callback('NewCallerId',
                                           channel_updater.parse_new_caller_id)

        callback_handler.register_callback('Hold', channel_updater.parse_hold)
        callback_handler.register_callback('Inherit', channel_updater.parse_inherit)

        call_receiver = context.get('call_receiver')
        callback_handler.register_callback('Newstate', call_receiver.handle_newstate)
        callback_handler.register_callback('Hangup', call_receiver.handle_hangup)
        callback_handler.register_callback('Dial', call_receiver.handle_dial)

    def _register_message_hooks(self):
        message_hook.add_hook([('function', 'updateconfig'),
                               ('listname', 'users')],
                              lambda x: funckey_manager.parse_update_user_config(self._funckey_manager, x))

    def _init_statistics_producers(self):
        self._statistics_producer_initializer.init_queue_statistics_producer(self._queue_statistics_producer)

    def _init_agent_availability(self):
        for agent_status in self._agent_client.get_agent_statuses():
            if agent_status.logged:
                agent_status_cti = AgentStatus.available
            else:
                agent_status_cti = AgentStatus.logged_out
            dao.agent.set_agent_availability(agent_status.id, agent_status_cti)
        context.get('agent_status_adapter').subscribe_all_logged_agents()

    def run(self):
        while True:
            try:
                self.main_loop()
            except Exception:
                logger.exception('main loop has crashed ... retrying in 5 seconds ...')
                time.sleep(5)

    def manage_tcp_connections(self, sel_i, msg, kind):
        """
        Dispatches the message's handling according to the connection's kind.
        """
        closemenow = False
        if not isinstance(kind, str):  # CTI, INFO, WEBI
            replies = kind.manage_connection(msg)
            for reply in replies:
                if reply:
                    if 'closemenow' in reply:
                        closemenow = reply.get('closemenow')
                    if 'message' in reply:
                        if reply.get('dest'):
                            self.send_to_cti_client(reply.get('dest'),
                                                    reply.get('message'))
                        else:
                            kind.reply(reply.get('message'))
                    elif 'warning' in reply:
                        kind.reply(reply.get('warning'))
        else:
            logger.warning('unknown connection kind %s', kind)
        return closemenow

    def checkqueue(self):
        ncount = 0
        while self.timeout_queue.qsize() > 0:
            ncount += 1
            (toload,) = self.timeout_queue.get()
            action = toload.get('action')
            if action == 'ctilogin':
                connc = toload.get('properties')
                connc.close()
                if connc in self.fdlist_established:
                    del self.fdlist_established[connc]
            else:
                logger.warning('checkqueue : unknown action received : %s', action)
        return ncount

    def cb_timer(self, *args):
        try:
            tname = threading.currentThread()
            tname.setName('Thread-main')
            self.timeout_queue.put(args)
            os.write(self.pipe_queued_threads[1], 'main:\n')
        except Exception:
            logger.exception('cb_timer %s', args)

    def main_loop(self):
        self.askedtoquit = False
        self.time_start = time.localtime()
        logger.info('STARTING %s (pid %d))', self.servername, os.getpid())
        logger.info('(1/3) Retrieving data')

        self.safe = context.get('innerdata')
        self.safe.user_service_manager = self._user_service_manager
        self.safe.queue_member_cti_adapter = self._queue_member_cti_adapter

        dao.instanciate_dao(self.safe, self._queue_member_manager)

        self.safe.init_xod_config()
        self.safe.init_xod_status()
        self.safe.register_cti_handlers()
        self.safe.register_ami_handlers()
        self.safe.update_directories()

        self._queue_member_updater.on_initialization()
        self._queue_member_cti_subscriber.send_cti_event = self.send_cti_event
        self._queue_member_cti_subscriber.subscribe_to_queue_member(self._queue_member_notifier)
        self._queue_member_indexer.subscribe_to_queue_member(self._queue_member_notifier)
        self._queue_statistics_manager.subscribe_to_queue_member(self._queue_member_notifier)
        self._queue_statistics_producer.subscribe_to_queue_member(self._queue_member_notifier)
        self._init_statistics_producers()
        self._init_agent_availability()
        self._queue_member_indexer.initialize(self._queue_member_manager)

        logger.info('(2/3) Local AMI socket connection')
        self.interface_ami.init_connection()
        self.ami_sock = self.interface_ami.connect()
        if not self.ami_sock:
            self._on_ami_down()

        logger.info('(3/3) Listening sockets')
        xivoconf_general = self._config.getconfig('main')
        socktimeout = float(xivoconf_general.get('sockettimeout', '2'))
        socket.setdefaulttimeout(socktimeout)

        self.fdlist_established = {}
        self.fdlist_listen_cti = {}

        incoming_tcp = xivoconf_general.get('incoming_tcp', {})
        for kind, bind_and_port in incoming_tcp.iteritems():
            allow_kind = True
            if len(bind_and_port) > 2:
                allow_kind = bind_and_port[2]
            if not allow_kind:
                logger.warning('%s kind listening socket has been explicitly disabled', kind)
                continue
            bind, port = bind_and_port[:2]
            self._init_tcp_socket(kind, bind, port)

        logger.info('CTI Fully Booted in %.6f seconds', (time.time() - self.start_time))
        while not self.askedtoquit:
            self.select_step()

    def _init_tcp_socket(self, kind, bind, port):
        try:
            trueport = int(port) + cti_config.PORTDELTA
            gai = socket.getaddrinfo(bind, trueport, 0, socket.SOCK_STREAM, socket.SOL_TCP)
            if not gai:
                return
            (afinet, socktype, proto, dummy, bindtuple) = gai[0]
            UIsock = socket.socket(afinet, socktype)
            UIsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            UIsock.bind(bindtuple)
            UIsock.listen(10)
            self.fdlist_listen_cti[UIsock] = '%s:%s' % (kind, 1)
        except Exception:
            logger.exception('tcp %s %d', bind, trueport)

    def _on_ami_down(self):
        logger.warning('AMI: CLOSING (%s)', time.asctime())
        logger.info('shutting down xivo-ctid')
        sys.exit(2)

    def get_connected(self, tomatch):
        clist = list()
        for interface_obj in self.fdlist_established.itervalues():
            if not isinstance(interface_obj, str) and interface_obj.kind in ['CTI', 'CTIS']:
                userid = interface_obj.connection_details.get('userid')
                if self.safe.user_match(userid, tomatch):
                    clist.append(interface_obj)
        return clist

    def sendsheettolist(self, tsl, payload):
        for k in tsl:
            k.reply(payload)

    def send_cti_event(self, event):
        self._cti_events.put(event)
        message_hook.run_hooks(event)

    def _empty_cti_events_queue(self):
        while self._cti_events.qsize() > 0:
            msg = self._cti_events.get()
            for interface_obj in self.fdlist_established.itervalues():
                if not isinstance(interface_obj, str) and interface_obj.kind in ['CTI', 'CTIS']:
                    interface_obj.append_msg(msg)

    def set_transfer_socket(self, faxobj, direction):
        for iconn, kind in self.fdlist_established.iteritems():
            if kind.kind in ['CTI', 'CTIS']:
                peername = '%s:%d' % iconn.getpeername()
                if peername == faxobj.socketref:
                    kind.set_as_transfer(direction, faxobj)
                    if direction == 's2c':
                        sendbuffer = ''
                        kind.reply(sendbuffer)
                    break

    def send_to_cti_client(self, who, what):
        (ipbxid, userid) = who.split('/')
        for interface_obj in self.fdlist_established.itervalues():
            if not isinstance(interface_obj, str) and interface_obj.kind in ['CTI', 'CTIS']:
                if interface_obj.connection_details.get('userid') == userid:
                    interface_obj.reply(what)

    def _init_socket(self):
        try:
            fdtodel = []
            for cn in self.fdlist_established:
                if isinstance(cn, ClientConnection):
                    if cn.isClosed and cn not in fdtodel:
                        fdtodel.append(cn)
                    if cn.toClose and not cn.need_sending():
                        cn.close()
                        if cn not in fdtodel:
                            fdtodel.append(cn)
            for cn in fdtodel:
                del self.fdlist_established[cn]

            self.fdlist_full = []
            self.fdlist_full.append(self.pipe_queued_threads[0])
            self.fdlist_full.append(self.ami_sock)
            self.fdlist_full.extend(self.fdlist_listen_cti)
            self.fdlist_full.extend(self.fdlist_established)

            writefds = []
            for iconn, kind in self.fdlist_established.iteritems():
                if kind.kind in ['CTI', 'CTIS'] and iconn.need_sending():
                    writefds.append(iconn)
            sels_i, sels_o, sels_e = select.select(self.fdlist_full, writefds, [])
            return (sels_i, sels_o, sels_e)

        except Exception:
            logger.exception('(select) probably Ctrl-C or daemon stop or daemon restart ...')
            logger.warning('(select) self.askedtoquit=%s fdlist_full=%s', self.askedtoquit, self.fdlist_full)
            logger.warning('(select) current open TCP connections : (CTI, WEBI, INFO) %s', self.fdlist_established)
            logger.warning('(select) current open TCP connections : (AMI) %s', self.ami_sock)

            self._socket_close_all()

            if self.askedtoquit:
                time_uptime = int(time.time() - time.mktime(self.time_start))
                logger.info('STOPPING %s (pid %d) / uptime %d s (since %s)',
                            self.servername, os.getpid(),
                            time_uptime, time.asctime(self.time_start))
                for t in filter(lambda x: x.getName() != 'MainThread', threading.enumerate()):
                    print '--- (stop) killing thread <%s>' % t.getName()
                    t._Thread__stop()
                daemonize.unlock_pidfile(cti_config.PIDFILE)
                sys.exit(5)
            else:
                self.askedtoquit = True
                for t in filter(lambda x: x.getName() != 'MainThread', threading.enumerate()):
                    print '--- (reload) the thread <%s> remains' % t.getName()
                return (None, None, None)

    def _socket_close_all(self):
        for s in self.fdlist_full:
            if s in self.fdlist_established:
                if self.askedtoquit:
                    self.fdlist_established[s].disconnected(DisconnectCause.by_server_stop)
                else:
                    self.fdlist_established[s].disconnected(DisconnectCause.by_server_reload)
            if not isinstance(s, int):
                s.close()

    def _socket_ami_read(self, sel_i):
        buf = sel_i.recv(cti_config.BUFSIZE_LARGE)
        if not buf:
            self._on_ami_down()
        else:
            self.interface_ami.handle_event(buf)

    def _socket_detect_new_tcp_connection(self, sel_i):
        [kind, nmax] = self.fdlist_listen_cti[sel_i].split(':')
        [socketobject, address] = sel_i.accept()

        ctiseparator = '\n'
        if kind == 'CTI':
            socketobject = ClientConnection(socketobject, address, ctiseparator)
            interface = interface_cti.CTI(self)
        elif kind == 'CTIS':
            certfile = self._config.getconfig('main')['certfile']
            keyfile = self._config.getconfig('main')['keyfile']
            try:
                connstream = ssl.wrap_socket(socketobject,
                                             server_side=True,
                                             certfile=certfile,
                                             keyfile=keyfile,
                                             ssl_version=cti_config.SSLPROTO)
                socketobject = ClientConnection(connstream, address, ctiseparator)
                interface = interface_cti.CTIS(self)
            except ssl.SSLError:
                logger.exception('%s:%s:%d cert=%s key=%s)',
                                 kind, address[0], address[1],
                                 certfile,
                                 keyfile)
                socketobject.close()
                socketobject = None

        if socketobject:
            if kind in ['CTI', 'CTIS']:
                logintimeout = int(self._config.getconfig('main').get('logintimeout', 5))
                interface.logintimer = threading.Timer(logintimeout, self.cb_timer,
                                                       ({'action': 'ctilogin',
                                                       'properties': socketobject},))
                interface.logintimer.start()
            elif kind == 'INFO':
                interface = interface_info.INFO(self)
            elif kind == 'WEBI':
                interface = interface_webi.WEBI(self, self._queue_member_updater)

            interface.connected(socketobject)

            self.fdlist_established[socketobject] = interface
        else:
            logger.warning('socketobject is not defined ...')

    def _socket_established_read(self, sel_i):
        try:
            interface_obj = self.fdlist_established[sel_i]
            requester = '%s:%d' % sel_i.getpeername()[:2]
            closemenow = False
            if isinstance(sel_i, ClientConnection):
                try:
                    lines = sel_i.readlines()
                    for line in lines:
                        if line:
                            closemenow = self.manage_tcp_connections(sel_i, line, interface_obj)
                except ClientConnection.CloseException:
                    interface_obj.disconnected(DisconnectCause.broken_pipe)
                    del self.fdlist_established[sel_i]
            else:
                try:
                    msg = sel_i.recv(cti_config.BUFSIZE_LARGE, socket.MSG_DONTWAIT)
                except socket.error:
                    logger.exception('connection to %s (%s)', requester, interface_obj)
                    msg = ''

                if msg:
                    try:
                        closemenow = self.manage_tcp_connections(sel_i, msg, interface_obj)
                    except Exception:
                        logger.exception('handling %s (%s)', requester, interface_obj)
                else:
                    closemenow = True

            if closemenow:
                interface_obj.disconnected(DisconnectCause.by_client)
                sel_i.close()
                del self.fdlist_established[sel_i]
        except Exception:
            logger.exception('[%s] %s', interface_obj, sel_i)
            logger.warning('unexpected socket breakup')
            interface_obj.disconnected(DisconnectCause.broken_pipe)
            sel_i.close()
            del self.fdlist_established[sel_i]

    def _socket_pipe_queue_read(self, sel_i):
        # try:
        pipebuf = os.read(sel_i, 1024)
        if not pipebuf:
            logger.warning('pipe_queued_threads has been closed')
        else:
            for pb in pipebuf.split('\n'):
                if not pb:
                    continue
                [kind, where] = pb.split(':')
                if kind in ['main', 'innerdata', 'ami']:
                    if kind == 'main':
                        self.checkqueue()
                    elif kind == 'innerdata':
                        self.safe.checkqueue()
                    elif kind == 'ami':
                        self.interface_ami.checkqueue()
                else:
                    logger.warning('unknown kind for %s', pb)
        # except Exception:
        #    logger.exception('[pipe_queued_threads]')

    def _update_safe_list(self):
        if self.update_config_list:
            try:
                if 'xivo[cticonfig,update]' in self.update_config_list:
                    self._config.update()
                    self.safe.update_directories()
                    self.update_config_list.pop(self.update_config_list.index('xivo[cticonfig,update]'))
            except Exception:
                logger.exception('failed while executing xivo[cticonfig,update]')
            try:
                while self.update_config_list:
                    msg = self.update_config_list.pop()
                    self.safe.update_config_list('%ss' % msg['object_name'], msg['state'], msg['id'])
                    self._empty_cti_events_queue()
            except Exception:
                logger.exception('Config reload (computed timeout)')

    def select_step(self):
        sels_i, sels_o, sels_e = self._init_socket()

        try:
            for sel_o in sels_o:
                try:
                    sel_o.process_sending()
                except ClientConnection.CloseException:
                    if sel_o in self.fdlist_established:
                        kind = self.fdlist_established[sel_o]
                        kind.disconnected(DisconnectCause.broken_pipe)
                        sel_o.close()
                        del self.fdlist_established[sel_o]
        except Exception:
            logger.exception('Socket writer')

        try:
            for sel_i in sels_i:
                # these AMI connection are used in order to manage AMI commands and events
                if sel_i == self.ami_sock:
                    self._socket_ami_read(sel_i)
                # the new TCP connections (CTI, WEBI, INFO) are catched here
                elif sel_i in self.fdlist_listen_cti:
                    self._socket_detect_new_tcp_connection(sel_i)
                # incoming TCP connections (CTI, WEBI, INFO)
                elif sel_i in self.fdlist_established:
                    self._socket_established_read(sel_i)
                # local pipe fd
                elif self.pipe_queued_threads[0] == sel_i:
                    self._socket_pipe_queue_read(sel_i)

                self._update_safe_list()
                self._empty_cti_events_queue()
        except Exception:
            logger.exception('Socket Reader')
